#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan@google.com>
#
# SPDX-License-Identifier: Apache-2.0

"Project description variables"

PROJECT_NAME = "kapitan"
VERSION = '0.27.3-rc.2'
DESCRIPTION = "Generic templated configuration management for Kubernetes, " "Terraform and other things"
AUTHOR = "Ricardo Amaro"
AUTHOR_EMAIL = "ramaro@google.com"
LICENCE = "Apache License 2.0"
URL = "https://github.com/deepmind/kapitan"